from .merging_problem import MergingProblem, MultiObjectiveMergingProblem
from .predefined_problems import (
    CrossLingualMathProblem,
    LmEvalMultiObjectiveProblem,
    MultilingualMergingProblem,
)
